# balpy_v2/lib/__init__.py
from enum import Enum


class Chain(Enum):
    mainnet = 1
    polygon = 137
    arbitrum = 42161
    gnosis = 100
    optimism = 10
    goerli = 5

    def __init__(self, id):
        self.id = id



from unittest import result
import pandas as pd
from balpy_v2.lib.flatten_json import flatten_json
from balpy_v2.lib.gql import gql

GRAPH_URL = "https://api.thegraph.com/subgraphs/name/bleu-studio/balancer-mainnet-v2"
QUERY = """
query token($block: Int!) {
    poolTokens(block: {number: $block},
               first: 1000,
               where:{
                    paidProtocolFees_not:"0"
               },
                orderBy: paidProtocolFees,
                orderDirection: desc
        ){
    id
    symbol
    address
    paidProtocolFees
    token {
      latestUSDPrice
    }
	pool: poolId {
      id
      address
      symbol
    }
  }
}
"""

import datetime
import asyncio
from balpy_v2.subgraphs.blocks import get_block_number_by_timestamp

REPORT_PERIOD_START_DATE = datetime.datetime(2022, 7, 11, 0, 0, 0).timestamp()

# 2 weeks in seconds
NOW = datetime.datetime.utcnow().timestamp()
REPORT_PERIOD_DURATION = datetime.timedelta(weeks=2).total_seconds()

import math

REPORT_CYCLES_UNTIL_NOW = math.ceil(
    (NOW - REPORT_PERIOD_START_DATE) / REPORT_PERIOD_DURATION
)

REPORT_CYCLES_START_DATES = [
    REPORT_PERIOD_START_DATE + REPORT_PERIOD_DURATION * i
    for i in range(REPORT_CYCLES_UNTIL_NOW)
]

CYCLES_NAMES_FROM_DATES = [
    f"""{datetime.datetime.fromtimestamp(REPORT_PERIOD_START_DATE + REPORT_PERIOD_DURATION * i).strftime('%Y.%m.%d')}\
    -\
    {datetime.datetime.fromtimestamp(REPORT_PERIOD_START_DATE + REPORT_PERIOD_DURATION * (i + 1)).strftime('%Y.%m.%d')}""".replace(
        " ", ""
    )
    for i in range(REPORT_CYCLES_UNTIL_NOW)
]


async def report_cycles_block_numbers():
    return await asyncio.gather(
        *[
            get_block_number_by_timestamp(timestamp=int(ts))
            for ts in REPORT_CYCLES_START_DATES
        ]
    )


async def report_cycles_data():
    blocks = await report_cycles_block_numbers()
    # blocks = list(filter(lambda x: x < 16253377, blocks))
    results = await asyncio.gather(
        *[
            gql(
                GRAPH_URL,
                QUERY,
                variables=dict(block=block),
            )
            for block in blocks
        ]
    )

    return [
        (result, blocks[idx], REPORT_CYCLES_START_DATES[idx])
        for idx, result in enumerate(results)
    ]


def period_analysis(start, end):
    period_start_data, start_block, start_date = start
    period_end_data, end_block, end_date = end
    period_start_df = (
        pd.DataFrame([flatten_json(x) for x in period_start_data["poolTokens"]])
        .set_index(["id"])
        .add_suffix("_start")
    )
    period_start_df["block_start"] = start_block
    period_start_df["date_start"] = start_date
    period_end_df = (
        pd.DataFrame([flatten_json(x) for x in period_end_data["poolTokens"]])
        .set_index("id")
        .add_suffix("_end")
    )
    period_end_df["block_end"] = end_block
    period_end_df["date_end"] = end_date
    merged = period_start_df.merge(period_end_df, left_index=True, right_index=True)
    # drop duplicated columns
    merged = (
        merged.loc[:, ~merged.columns.duplicated()]
        .copy()
        .set_index("date_start", append=True)
    )
    merged["address_start"] = "ethereum:" + merged["address_start"]
    merged[
        [
            "token_latestUSDPrice_end",
            "token_latestUSDPrice_start",
            "paidProtocolFees_end",
            "paidProtocolFees_start",
        ]
    ] = merged[
        [
            "token_latestUSDPrice_end",
            "token_latestUSDPrice_start",
            "paidProtocolFees_end",
            "paidProtocolFees_start",
        ]
    ].astype(
        float
    )
    merged["paidProtocolFees_diff"] = (
        merged["paidProtocolFees_end"] - merged["paidProtocolFees_start"]
    )
    paid_fees = merged[merged["paidProtocolFees_diff"] != 0].sort_values(
        "paidProtocolFees_diff", ascending=False
    )
    paid_fees["paidProtocolFees_inUSD_end"] = (
        paid_fees["paidProtocolFees_diff"] * paid_fees["token_latestUSDPrice_end"]
    )
    paid_fees["paidProtocolFees_inUSD_start"] = (
        paid_fees["paidProtocolFees_diff"] * paid_fees["token_latestUSDPrice_start"]
    )
    return paid_fees


async def full_analysis():
    data = await report_cycles_data()
    cycles_data = list(zip(data, data[1:] + data[:1]))[:-1]
    return pd.concat([period_analysis(*x) for x in cycles_data])



import httpx

API_BASE_URL = "https://coins.llama.fi"


async def base_request(path):
    async with httpx.AsyncClient() as client:
        r = await client.get(API_BASE_URL + path)
    return r.json()


async def get_historical_prices(timestamp, coins):
    path = "/prices/historical/{timestamp}/{coins}"
    return await base_request(path.format(timestamp=timestamp, coins=coins))


async def get_current_prices(coins):
    path = "/prices/current/{coins}"
    return await base_request(path.format(coins=coins))



from abc import ABC, abstractmethod

from balpy_v2.lib import Chain
from balpy_v2.subgraphs.client import GraphQLClient


class GraphQLQuery(ABC):
    def __init__(self, chain=Chain.mainnet, variables=dict()):
        self.chain = chain
        self.variables = variables

    @abstractmethod
    def get_query(self):
        pass

    @abstractmethod
    def get_client(self) -> GraphQLClient:
        pass

    async def execute(self):
        query = self.get_query()
        client = self.get_client()
        return await client.__class__.query(self.chain, query, self.variables)



from balpy_v2.lib import Chain
from balpy_v2.subgraphs.client import SubgraphBaseClient
from balpy_v2.subgraphs.query import GraphQLQuery

BASE_URL = "https://api.thegraph.com/subgraphs/name/balancer-labs"

BALANCER_MAINNET_SUBGRAPH_URL_MAP = {
    Chain.mainnet: BASE_URL + "/balancer-v2",
    Chain.polygon: BASE_URL + "/balancer-polygon-v2",
    Chain.arbitrum: BASE_URL + "/balancer-arbitrum-v2",
    Chain.gnosis: BASE_URL + "/balancer-gnosis-chain-v2",
    # TODO: missing subgraph link in https://docs.balancer.fi/reference/subgraph/
    # optimism="",
    # avalanche="",
}


class BalancerSubgraph(SubgraphBaseClient):
    def get_url(self, chain):
        return BALANCER_MAINNET_SUBGRAPH_URL_MAP[chain]


class BalancerSubgraphQuery(GraphQLQuery):
    def get_client(self):
        return BalancerSubgraph(self.chain)



# bots/lib/gql.py
import json

import httpx


class GraphQLError(Exception):
    pass


async def gql(url, query, variables={}):
    async with httpx.AsyncClient() as client:
        r = await client.post(
            url,
            json=dict(query=query, variables=variables),
        )

    try:
        return r.json()["data"]
    except KeyError:
        print(json.dumps(r.json(), indent=2))
        raise GraphQLError



from balpy_v2.lib import Chain
from balpy_v2.subgraphs.balancer import (
    BALANCER_MAINNET_SUBGRAPH_URL_MAP,
    BalancerSubgraph,
)
from balpy_v2.lib.gql import gql
from balpy_v2.subgraphs.balancer import BalancerSubgraphQuery


class TokensQuery(BalancerSubgraphQuery):
    def get_query(self):
        return """
            query($address_list:[String!]) {
              tokens(
                where:{address_in:$address_list}
              ) {
                id
                latestUSDPrice
              }
            }"""


async def async_token_query(chain=Chain.mainnet, address_list=[]):
    response = await TokensQuery(
        chain=chain, variables=dict(address_list=address_list)
    ).execute()
    return {token["id"]: token["latestUSDPrice"] for token in response["tokens"]}



import asyncio

from balpy_v2.lib import Chain
from balpy_v2.subgraphs.balancer import BALANCER_MAINNET_SUBGRAPH_URL_MAP
from balpy_v2.lib.gql import gql


from abc import ABC, abstractmethod


class GraphQLClient(ABC):
    def __init__(self, chain):
        self.url = self.get_url(chain)

    async def instance_query(self, query, variables=dict()):
        return await gql(self.url, query, variables=variables)

    @abstractmethod
    def get_url(self, chain):
        pass

    @classmethod
    async def query(cls, chain=Chain.mainnet, query=None, variables=dict()):
        if not query:
            raise ValueError("query must be provided")

        client = cls(chain)
        return await client.instance_query(query, variables)



# bots/lib/balancer/subgraphs/pools.py
import asyncio

from balpy_v2.lib import Chain
from balpy_v2.subgraphs.balancer import BALANCER_MAINNET_SUBGRAPH_URL_MAP
from balpy_v2.subgraphs.blocks import get_block_number_by_timestamp
from balpy_v2.lib.gql import gql
from balpy_v2.subgraphs.client import (
    BalancerSubgraph,
    SubgraphBaseClient,
    SubgraphClient,
)
from balpy_v2.subgraphs.balancer import BalancerSubgraphQuery


class BalancersQuery(BalancerSubgraphQuery):
    def get_query(self):
        return """
query($block: Block_height) {
  balancers(block: $block) {
    totalLiquidity
    totalSwapVolume
  }
}
"""


async def get_total_liquidity(chain=Chain.mainnet):
    r = await BalancersQuery(chain=chain).execute()
    return float(r["balancers"][0]["totalLiquidity"])


async def get_current_cross_chain_liquidity():
    return sum(
        [
            await get_total_liquidity(chain)
            for chain in BALANCER_MAINNET_SUBGRAPH_URL_MAP.keys()
        ]
    )


async def get_current_total_swap_volume(chain=Chain.mainnet):
    r = await BalancersQuery(chain=chain).execute()
    return float(r["balancers"][0]["totalSwapVolume"])


async def get_t_minus_24h_total_swap_volume(chain=Chain.mainnet):
    t_minus_24h_block_number = await get_block_number_by_timestamp(chain=chain)
    r = await BalancersQuery(
        chain=chain, variables=dict(block=dict(number=t_minus_24h_block_number))
    ).execute()
    return float(r["balancers"][0]["totalSwapVolume"])


async def get_last_24h_total_swap_volume(chain=Chain.mainnet):
    return await get_current_total_swap_volume(
        chain=chain
    ) - await get_t_minus_24h_total_swap_volume(chain=chain)


async def get_cross_chain_last_24h_total_swap_volume():
    chains = BALANCER_MAINNET_SUBGRAPH_URL_MAP.keys()
    asyncs = await asyncio.gather(
        *[get_last_24h_total_swap_volume(chain=chain) for chain in chains]
    )

    return sum(asyncs)



from time import time

MINUTE_IN_SECONDS = 60
HOUR_IN_SECONDS = MINUTE_IN_SECONDS * 60
WEEK_IN_SECONDS = HOUR_IN_SECONDS * 24 * 7
YEAR_IN_SECONDS = HOUR_IN_SECONDS * 24 * 365


def get_time_1h_ago(time=time()):
    return int(time) - HOUR_IN_SECONDS


def get_time_1w_later(time=time()):
    return int(time) + WEEK_IN_SECONDS


def get_time_1w_ago(time=time()):
    return int(time) - WEEK_IN_SECONDS


def get_timestamps(time=time()):
    def to_str(x):
        return str(x)

    return dict(
        timestamp_gte=to_str(time), timestamp_lt=to_str(time + (MINUTE_IN_SECONDS * 10))
    )


def get_time_24h_ago(time=time()):
    return int(time - (HOUR_IN_SECONDS * 24))


def get_time_1w_ago(time=time()):
    return int(time - WEEK_IN_SECONDS)



# balpy_v2/lib/blocks/subgraph.py
import os

import httpx
from async_lru import alru_cache

from balpy_v2.lib import Chain
from balpy_v2.lib.gql import gql
from balpy_v2.lib.time import get_time_24h_ago, get_timestamps

BLOCKS_SUBGRAPH_URL_MAP = {
    Chain.mainnet: "https://api.thegraph.com/subgraphs/name/blocklytics/ethereum-blocks",
    Chain.polygon: "https://api.thegraph.com/subgraphs/name/ianlapham/polygon-blocks",
    Chain.arbitrum: "https://api.thegraph.com/subgraphs/name/ianlapham/arbitrum-one-blocks",
}


CHAIN_BLOCK_EXPLORER_FN_MAP = {
    Chain.gnosis: lambda timestamp: f"https://api.gnosisscan.io/api?module=block&action=getblocknobytime&timestamp={timestamp}&closest=before&apikey={os.getenv('GNOSIS_API_KEY')}",
}

BLOCKS_QUERY = """
query($timestamp_gte: BigInt, $timestamp_lt: BigInt) {
	blocks(
        first: 1,
        orderBy: number,
        orderDirection: asc,
        where: {
            timestamp_gte: $timestamp_gte,
            timestamp_lt: $timestamp_lt
        }
      ) {
            number
      }
}
"""


CHAIN_AVG_BLOCK_TIME = {
    Chain.mainnet: 13,
    Chain.polygon: 2,
    Chain.arbitrum: 2,
    Chain.gnosis: 2,
}


async def best_guess(chain=Chain.gnosis, t=get_time_24h_ago()) -> int:
    async with httpx.AsyncClient() as client:
        r = await client.get(CHAIN_BLOCK_EXPLORER_FN_MAP[chain](t))

    return int(r.json()["result"])


@alru_cache
async def get_block_number_by_timestamp(
    chain=Chain.mainnet, timestamp=get_time_24h_ago()
) -> int:
    url = BLOCKS_SUBGRAPH_URL_MAP.get(chain)

    if not url:
        return await best_guess(chain, timestamp)

    data = await gql(
        BLOCKS_SUBGRAPH_URL_MAP[chain],
        BLOCKS_QUERY,
        variables=get_timestamps(timestamp),
    )
    return int(data["blocks"][0]["number"])



from abc import ABC, abstractmethod

from balpy_v2.lib import Chain
from balpy_v2.lib.gql import gql
from balpy_v2.subgraphs.balancer import BalancerSubgraph
from balpy_v2.subgraphs.client import SubgraphBaseClient
from balpy_v2.subgraphs.balancer import BalancerSubgraphQuery


class PoolQuery(BalancerSubgraphQuery):
    def get_query(self):
        return """
        query($id_list:[ID!]){
          pools(where:{id_in:$id_list}) {
            id
            symbol
          }
        }
        """


async def pool_query(chain=Chain.mainnet, pool_id_list=[]):
    response = await PoolQuery(
        chain=chain, variables={"id_list": pool_id_list}
    ).execute()
    return {pool["id"]: pool["symbol"] for pool in response["pools"]}



def flatten_json(nested_json):
    """
    Flatten json object with nested keys into a single level.
    Args:
        nested_json: A nested json object.
    Returns:
        The flattened json object if successful, None otherwise.
    """
    out = {}

    def flatten(x, name=""):
        if type(x) is dict:
            for a in x:
                flatten(x[a], name + a + "_")
        elif type(x) is list:
            i = 0
            for a in x:
                flatten(a, name + str(i) + "_")
                i += 1
        else:
            out[name[:-1]] = x

    flatten(nested_json)
    return out
